import random

from geoformat_lib.explore_data.random_geometry import (
    random_point,
    random_segment,
    random_bbox
)
from geoformat_lib.geoprocessing.connectors.operations import segment_to_bbox
from geoformat_lib.geoprocessing.connectors.predicates import (
    point_intersects_bbox,
    bbox_intersects_bbox
)
from geoformat_lib.index.geometry.grid import (
    bbox_to_g_id,
    point_to_g_id,
    g_id_to_bbox,
    g_id_to_point,
    g_id_neighbor_in_grid_index,
    create_grid_index,
    grid_index_to_geolayer
)
from tests.data.geolayers import (
    geolayer_for_index,
    geolayer_for_index_grid_geolayer,
    geolayer_for_index_width,
    geolayer_for_index_width_grid_geolayer,
    geolayer_for_index_height,
    geolayer_for_index_height_grid_geolayer,
    geolayer_for_index_width_height,
    geolayer_for_index_width_height_grid_geolayer
)

from tests.data.geometry_index import (
    geolayer_for_index_grid_index,
    geolayer_for_index_width_grid_index,
    geolayer_for_index_height_grid_index,
    geolayer_for_index_width_height_grid_index
)

from tests.test_all import test_function

bbox_to_g_id_parameters = {
    0: {
        "bbox": (-9, -9, 9, 9),
        "mesh_size": 10,
        "x_grid_origin": 0.,
        "y_grid_origin": 0.,
        "grid_precision": None,
        "return_value": ((-1, -1), (-1, 0), (0, -1), (0, 0),)
    },
    1: {
        "bbox": (-9, -9, 9, 9),
        "mesh_size": 10,
        "x_grid_origin": 9,
        "y_grid_origin": 9,
        "grid_precision": None,
        "return_value": ((-2, -2), (-2, -1), (-2, 0), (-1, -2), (-1, -1), (-1, 0), (0, -2), (0, -1), (0, 0))
    },
    2: {
        "bbox": (2, 1, 5, 3.3333332),
        "mesh_size": 10/3,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "grid_precision": None,
        "return_value": ((0, 0), (1, 0))
    },
    3: {
        "bbox": (2, 1, 5, 3.3333332),
        "mesh_size": 10/3,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "grid_precision": 2,
        "return_value": ((0, 0), (0, 1), (1, 0), (1, 1))
    },
}

point_to_g_id_parameters = {
    0: {
        "point": (168158.34003543295, 5388929.476522685),
        "mesh_size": 30,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((5605, 179630),),
    },
    1: {
        "point": (10, 10),
        "mesh_size": 10,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((0, 0), (0, 1), (1, 0), (1, 1)),
    },
    2: {
        "point": (0, 0),
        "mesh_size": 10,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((-1, -1), (-1, 0), (0, -1), (0, 0)),
    },
    3: {
        "point": (0, 10),
        "mesh_size": 10,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((-1, 0), (-1, 1), (0, 0), (0, 1)),
    },
    4: {
        "point": (10, 0),
        "mesh_size": 10,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((0, -1), (0, 0), (1, -1), (1, 0)),
    },
    5: {
        "point": (5, 10),
        "mesh_size": 10,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((0, 0), (0, 1)),
    },
    6: {
        "point": (5, 0),
        "mesh_size": 10,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((0, -1), (0, 0)),
    },
    7: {
        "point": (0, 5),
        "mesh_size": 10,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((-1, 0), (0, 0)),
    },
    8: {
        "point": (10, 5),
        "mesh_size": 10,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((0, 0), (1, 0)),
    },
    9: {
        "point": (5, 5),
        "mesh_size": 10,
        "x_grid_origin": 5,
        "y_grid_origin": 5,
        "return_value": ((-1, -1), (-1, 0), (0, -1), (0, 0)),
    },
    10: {
        "point": (-76226, 5597112),
        "mesh_size": 30,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((-2541, 186570),)
    },
    11: {
        "point": (-33.07, 28.2),
        "mesh_size": 2.35,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "grid_precision": None,
        "return_value": ((-15, 12),)
    },
    12: {
        "point": (-33.07, 28.2),
        "mesh_size": 2.35,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "grid_precision": 8,
        "return_value": ((-15, 11), (-15, 12))
    },
    13: {
        "point": (0, 0),
        "mesh_size": 2,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((-1, -1), (-1, 0), (0, -1), (0, 0))
    },
    14: {
        "point": (0, 1),
        "mesh_size": 2,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((-1, 0), (0, 0))
    },
    15: {
        "point": (1, 1),
        "mesh_size": 2,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((0, 0),)
    },
    16: {
        "point": (1, 0),
        "mesh_size": 2,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((0, -1), (0, 0))
    },
    17: {
        "point": (1, -1),
        "mesh_size": 2,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((0, -1),)
    },
    18: {
        "point": (0, -1),
        "mesh_size": 2,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((-1, -1), (0, -1))
    },
    19: {
        "point": (-1, -1),
        "mesh_size": 2,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((-1, -1),)
    },
    20: {
        "point": (-1, 0),
        "mesh_size": 2,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((-1, -1), (-1, 0))
    },
    21: {
        "point": (-1, 1),
        "mesh_size": 2,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "return_value": ((-1, 0),)
    }
}

g_id_to_bbox_parameters = {
    0: {
        "g_id": (0, 0),
        "mesh_size": 10,
        "x_grid_origin": 0.,
        "y_grid_origin": 0.,
        "grid_precision": None,
        "return_value": (0, 0, 10, 10)
    },
    1: {
        "g_id": (0, 0),
        "mesh_size": 10,
        "x_grid_origin": 10,
        "y_grid_origin": -10,
        "grid_precision": None,
        "return_value": (10, -10, 20, 0)
    },
    2: {
        "g_id": (0, 0),
        "mesh_size": 3 / 9,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "grid_precision": 4,
        "return_value": (0, 0, 0.3333, 0.3333)
    },
    3: {
        "g_id": (0, 0),
        "mesh_size": 3 / 9,
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "grid_precision": 8,
        "return_value": (0, 0, 0.33333333, 0.33333333)
    },
}


g_id_to_point_parameters = {
    0: {
        "g_id": (0, 0),
        "mesh_size": 10,
        "position": "center",
        "x_grid_origin": 0.,
        "y_grid_origin": 0.,
        "grid_precision": None,
        "return_value": (5, 5)
    },
    1: {
        "g_id": (0, 0),
        "mesh_size": 10,
        "position": "center",
        "x_grid_origin": 10,
        "y_grid_origin": -10,
        "grid_precision": None,
        "return_value": (15, -5)
    },
    2: {
        "g_id": (0, 0),
        "mesh_size": 3 / 9,
        "position": "center",
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "grid_precision": 4,
        "return_value": (0.1667, 0.1667)
    },
    3: {
        "g_id": (0, 0),
        "mesh_size": 3 / 9,
        "position": "center",
        "x_grid_origin": 0,
        "y_grid_origin": 0,
        "grid_precision": 8,
        "return_value": (0.16666667, 0.16666667)
    },
    4: {
        "g_id": (-10, -10),
        "mesh_size": 10,
        "position": "center",
        "x_grid_origin": 0.,
        "y_grid_origin": 0.,
        "grid_precision": None,
        "return_value": (-95., -95.)
    },
    5: {
        "g_id": (-10, -10),
        "mesh_size": 10,
        "position": "NW",
        "x_grid_origin": 0.,
        "y_grid_origin": 0.,
        "grid_precision": None,
        "return_value": (-100, -90)
    },
    6: {
        "g_id": (-10, -10),
        "mesh_size": 10,
        "position": "NE",
        "x_grid_origin": 0.,
        "y_grid_origin": 0.,
        "grid_precision": None,
        "return_value": (-90, -90)
    },
    7: {
        "g_id": (-10, -10),
        "mesh_size": 10,
        "position": "SW",
        "x_grid_origin": 0.,
        "y_grid_origin": 0.,
        "grid_precision": None,
        "return_value": (-100, -100)
    },
    8: {
        "g_id": (-10, -10),
        "mesh_size": 10,
        "position": "SE",
        "x_grid_origin": 0.,
        "y_grid_origin": 0.,
        "grid_precision": None,
        "return_value": (-90, -100)
    },
}

g_id_neighbor_in_grid_index_parameters = {
    0: {
        "g_id": (-10, -3),
        "grid_index": create_grid_index(geolayer_for_index, mesh_size=1),
        "nb_mesh": 1,
        "g_id_include": True,
        "return_value": ((-11, -4), (-11, -3), (-10, -4), (-10, -3), (-9, -4), (-9, -3))
    },
    1: {
        "g_id": (-10, -3),
        "grid_index": create_grid_index(geolayer_for_index, mesh_size=1),
        "nb_mesh": 2,
        "g_id_include": True,
        "return_value": ((-12, -5), (-12, -4), (-12, -3), (-11, -5), (-11, -4), (-11, -3), (-10, -5), (-10, -4),
                         (-10, -3), (-9, -5), (-9, -4), (-9, -3), (-8, -5), (-8, -4), (-8, -3))
    },
    2: {
        "g_id": (-10, -3),
        "grid_index": create_grid_index(geolayer_for_index, mesh_size=1),
        "nb_mesh": 1,
        "g_id_include": False,
        "return_value": ((-11, -4), (-11, -3), (-10, -4), (-9, -4), (-9, -3))
    },
    3: {
        "g_id": (-10, -3),
        "grid_index": create_grid_index(geolayer_for_index, mesh_size=1),
        "nb_mesh": 2,
        "g_id_include": False,
        "return_value": ((-12, -5), (-12, -4), (-12, -3), (-11, -5), (-11, -4), (-11, -3), (-10, -5), (-10, -4),
                         (-9, -5), (-9, -4), (-9, -3), (-8, -5), (-8, -4), (-8, -3))
    },
}

create_grid_index_parameters = {
    0: {
        'geolayer': geolayer_for_index,
        'mesh_size':  30,
        'x_grid_origin': 0,
        'y_grid_origin':  0,
        'return_value': geolayer_for_index_grid_index
    },
    1:  {
        'geolayer': geolayer_for_index_width,
        'mesh_size':  30,
        'x_grid_origin': 0,
        'y_grid_origin':  0,
        'return_value': geolayer_for_index_width_grid_index
    },
    2: {
        'geolayer': geolayer_for_index_height,
        'mesh_size': 30,
        'x_grid_origin': 0,
        'y_grid_origin': 0,
        'return_value': geolayer_for_index_height_grid_index
    },
    3:  {
        'geolayer': geolayer_for_index_width_height,
        'mesh_size':  30,
        'x_grid_origin': 0,
        'y_grid_origin':  0,
        'return_value': geolayer_for_index_width_height_grid_index
    },
}

grid_index_to_geolayer_parameters = {
    0: {
        "grid_index": geolayer_for_index_grid_index,
        "name": 'grid_index',
        "crs": None,
        "features_serialize": False,
        "return_value": geolayer_for_index_grid_geolayer
    },
    1: {
        "grid_index": geolayer_for_index_width_grid_index,
        "name": 'grid_index',
        "crs": None,
        "features_serialize": False,
        "return_value": geolayer_for_index_width_grid_geolayer
    },
    2: {
        "grid_index": geolayer_for_index_height_grid_index,
        "name": 'grid_index',
        "crs": None,
        "features_serialize": False,
        "return_value": geolayer_for_index_height_grid_geolayer
    },
    3: {
        "grid_index": geolayer_for_index_width_height_grid_index,
        "name": 'grid_index',
        "crs": None,
        "features_serialize": False,
        "return_value": geolayer_for_index_width_height_grid_geolayer
    }
}


def test_alea_point_to_g_id_to_bbox(nb_iteration):
    cnt_loop = 0
    bbox = (-100, -100, 100, 100)
    nb_round = 2
    for i in range(nb_iteration):
        test_point = random_point(bbox=bbox, nb_round=nb_round)
        mesh_size = round(random.random() * 10, 2)
        # print(test_point, mesh_size)
        if mesh_size < 1:
            mesh_size = 1
        for g_id in point_to_g_id(test_point, mesh_size):
            test_bbox = g_id_to_bbox(g_id, mesh_size, grid_precision=nb_round + 1)
            cnt_loop += 1
            if not point_intersects_bbox(test_point, test_bbox):
                print(test_point, mesh_size)
                return False
    print('\tnb g_id tested', cnt_loop)
    return True


def test_alea_segment_to_g_id_to_bbox(nb_iteration):
    bbox = (-100, -100, 100, 100)
    cnt_loop = 0
    nb_round = 2
    for i in range(nb_iteration):
        segment = random_segment(bbox=bbox, nb_round=nb_round)
        mesh_size = round(random.random() * 100, 2)
        # print(test_point, mesh_size)
        if mesh_size < 1:
            mesh_size = 1.
        segment_bbox = segment_to_bbox(segment)
        for g_id in bbox_to_g_id(segment_bbox, mesh_size):
            test_bbox = g_id_to_bbox(g_id, mesh_size, grid_precision=nb_round + 1)
            cnt_loop += 1
            if not bbox_intersects_bbox(segment_bbox, test_bbox):
                print(segment, segment_bbox, test_bbox, g_id, mesh_size)
                return False
    print('\tnb g_id tested', cnt_loop)
    return True


def test_alea_bbox_to_g_id_to_bbox(nb_iteration):
    bbox = (-100, -100, 100, 100)
    cnt_loop = 0
    nb_round = 2
    for i in range(nb_iteration):
        test_bbox_a = random_bbox(bbox=bbox, nb_round=nb_round)
        mesh_size = round(random.random() * 100, 2)
        # print(test_point, mesh_size)
        if mesh_size < 1:
            mesh_size = 1.
        for g_id in bbox_to_g_id(test_bbox_a, mesh_size):
            test_bbox_b = g_id_to_bbox(g_id, mesh_size, grid_precision=nb_round + 1)
            cnt_loop += 1
            if not bbox_intersects_bbox(test_bbox_a, test_bbox_b):
                print(test_bbox_a, test_bbox_b, g_id, mesh_size)
                return False
    print('\tnb g_id tested', cnt_loop)
    return True


def test_all():
    # bbox_to_g_id
    print(test_function(bbox_to_g_id, bbox_to_g_id_parameters))

    # point_to_g_id
    print(test_function(point_to_g_id, point_to_g_id_parameters))

    # g_id_to_bbox
    print(test_function(g_id_to_bbox, g_id_to_bbox_parameters))

    # g_id_to_point
    print(test_function(g_id_to_point, g_id_to_point_parameters))

    # g_id_neighbor_in_grid_index
    print(test_function(g_id_neighbor_in_grid_index, g_id_neighbor_in_grid_index_parameters))

    #  create_grid_index
    print(test_function(create_grid_index, create_grid_index_parameters))

    # grid_index_to_geolayer
    print(test_function(grid_index_to_geolayer, grid_index_to_geolayer_parameters))


    iteration = 1000000
    alea_point_to_g_id_to_bbox = test_alea_point_to_g_id_to_bbox(nb_iteration=iteration)
    if alea_point_to_g_id_to_bbox:
        print('Func test point_to_g_id / g_id_to_bbox / point_intersect_bbox OK')
    else:
        print('Func test point_to_g_id / g_id_to_bbox / point_intersect_bbox NOK')

    iteration = 10000
    alea_bbox_to_g_id_to_bbox = test_alea_segment_to_g_id_to_bbox(nb_iteration=iteration)
    if alea_bbox_to_g_id_to_bbox:
        print('Func test segment_to_bbox / bbox_to_g_id / g_id_to_bbox / bbox_intersects_bbox OK')
    else:
        print('Func segment_to_bbox/ test bbox_to_g_id / g_id_to_bbox / bbox_intersects_bbox NOK')


    iteration = 10000
    alea_bbox_to_g_id_to_bbox = test_alea_bbox_to_g_id_to_bbox(nb_iteration=iteration)
    if alea_bbox_to_g_id_to_bbox:
        print('Func test bbox_to_g_id / g_id_to_bbox / bbox_intersects_bbox OK')
    else:
        print('Func test bbox_to_g_id / g_id_to_bbox / bbox_intersects_bbox NOK')


if __name__ == '__main__':
    test_all()



# coding: utf-8

# In general Geoformat is licensed under an MIT/X style license with the
# following terms:
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# Authors :
#   Guilhain Averlant
#   Eliette Catelin
#   Quentin Lecuire
#   Charlotte Montesinos Chevalley
#   Coralie Rabiniaux

import sys

# GEOFORMAT conf variables
from geoformat_lib.conf.driver_variable import (
    OGR_FORMAT_CSV,
    OGR_FORMAT_KML,
    OGR_FORMAT_POSTGRESQL,
    OGR_FORMAT_GEOJSON,
    OGR_FORMAT_XLSX,
    OGR_FORMAT_ESRI_SHAPEFILE,
    OGR_DRIVER_NAMES)

from geoformat_lib.conf.geometry_variable import GEOMETRY_TYPE, GEOMETRY_CODE_TO_GEOMETRY_TYPE, GEOFORMAT_GEOMETRY_TYPE

# predicates
from geoformat_lib.geoprocessing.connectors.predicates import (
    point_intersects_point,
    point_intersects_segment,
    point_intersects_bbox,
    segment_intersects_bbox,
    segment_intersects_segment,
    bbox_intersects_bbox,
    point_position_segment,
    ccw_or_cw_segments
)

# operations
from geoformat_lib.geoprocessing.connectors.operations import (
    coordinates_to_point,
    coordinates_to_segment,
    coordinates_to_bbox,
    segment_to_bbox
)

# Geoprocessing

# measure
# distance
from geoformat_lib.geoprocessing.measure.distance import (
    euclidean_distance,
    manhattan_distance,
    point_vs_segment_distance,
    euclidean_distance_point_vs_segment
)
# length
from geoformat_lib.geoprocessing.measure.length import segment_length
from geoformat_lib.geoprocessing.length import geometry_length

# area
from geoformat_lib.geoprocessing.measure.area import shoelace_formula
from geoformat_lib.geoprocessing.area import geometry_area

# matrix
from geoformat_lib.geoprocessing.matrix.adjacency import (
    create_adjacency_matrix,
    get_neighbor_i_feat,
    get_area_intersecting_neighbors_i_feat
)

# line merge
from geoformat_lib.geoprocessing.line_merge import line_merge

# merge geometries
from geoformat_lib.geoprocessing.merge_geometries import merge_geometries

# point_on_linestring
from geoformat_lib.geoprocessing.point_on_linestring import (
    point_at_a_distance_on_segment,
    points_on_linestring_distance
)

# split
from geoformat_lib.geoprocessing.split import (
    segment_split_by_point,
    linestring_split_by_point
)

# union
from geoformat_lib.geoprocessing.union import (
    union_by_split
)

# geoparameters -> lines
from geoformat_lib.geoprocessing.geoparameters.lines import (
    line_parameters,
    perpendicular_line_parameters_at_point,
    point_at_distance_with_line_parameters,
    crossing_point_from_lines_parameters
)

# boundaries
from geoformat_lib.geoprocessing.geoparameters import boundaries

# geoparameters -> bbox
from geoformat_lib.geoprocessing.geoparameters.bbox import (
    bbox_union,
    extent_bbox,
    point_bbox_position
)

# bbox conversion
from geoformat_lib.conversion.bbox_conversion import (
    envelope_to_bbox,
    bbox_to_envelope,
    bbox_extent_to_2d_bbox_extent,
    bbox_to_polygon_coordinates
)

# bytes conversion
from geoformat_lib.conversion.bytes_conversion import (
    int_to_4_bytes_integer,
    float_to_double_8_bytes_array,
    coordinates_list_to_bytes,
    double_8_bytes_to_float,
    integer_4_bytes_to_int
)

# coordinates conversion
from geoformat_lib.conversion.coordinates_conversion import (
    format_coordinates,
    coordinates_to_2d_coordinates
)

# features conversion
from geoformat_lib.conversion.feature_conversion import (
    feature_serialize,
    feature_deserialize,
    features_geometry_ref_scan,
    features_fields_type_scan,
    feature_list_to_geolayer,
    feature_filter_geometry,
    feature_filter_attributes,
    feature_filter,
    features_filter
)

# fields conversion
from geoformat_lib.conversion.fields_conversion import (
    update_field_index,
    recast_field,
    drop_field
)

# geolayer conversion
from geoformat_lib.conversion.geolayer_conversion import (
    multi_geometry_to_single_geometry_geolayer,
    geolayer_to_2d_geolayer,
    create_geolayer_from_i_feat_list,
    reproject_geolayer
)

# geometry conversion
from geoformat_lib.conversion.geometry_conversion import (
    geometry_type_to_2d_geometry_type,
    geometry_to_2d_geometry,
    geometry_to_geometry_collection,
    multi_geometry_to_single_geometry,
    ogr_geometry_to_geometry,
    geometry_to_ogr_geometry,
    geometry_to_wkb,
    wkb_to_geometry,
    wkt_to_geometry,
    geometry_to_wkt,
    force_rhr,
    geometry_to_bbox,
    reproject_geometry
)

# metadata conversion
from geoformat_lib.conversion.metadata_conversion import (
    geometries_scan_to_geometries_metadata,
    fields_scan_to_fields_metadata,
    reorder_metadata_field_index_after_field_drop
)

# precision tolerance conversion
from geoformat_lib.conversion.precision_tolerance_conversion import (
    deduce_rounding_value_from_float,
    deduce_precision_from_round
)

# segment conversion
from geoformat_lib.conversion.segment_conversion import (
    segment_list_to_linestring
)

# database operation
from geoformat_lib.db.db_request import (
    sql,
    sql_select_to_geolayer
)

# geometry conversion
from geoformat_lib.conversion.geometry_conversion import (
    bbox_extent_to_2d_bbox_extent,
    geometry_type_to_2d_geometry_type,
    geometry_to_2d_geometry,
    geometry_to_geometry_collection
)

# data printing
from geoformat_lib.explore_data.print_data import (
    get_features_data_table_line,
    print_features_data_table,
    get_fields_metadata_table_line,
    print_metadata_field_table
)
# random geometries
from geoformat_lib.explore_data.random_geometry import (
    random_point,
    random_segment,
    random_bbox
)
# attributes index
from geoformat_lib.index.attributes.hash import (
    create_attribute_index
)

# grid index
from geoformat_lib.index.geometry.grid import (
    bbox_to_g_id,
    point_to_g_id,
    g_id_to_bbox,
    g_id_to_point,
    g_id_neighbor_in_grid_index,
    create_grid_index,
    grid_index_to_geolayer
)

# driver
from geoformat_lib.driver.ogr.ogr_driver import (
    ogr_layer_to_geolayer,
    ogr_layers_to_geocontainer,
    geolayer_to_ogr_layer,
    geocontainer_to_ogr_data_source
)

# clauses
from geoformat_lib.processing.data.clauses import (
    clause_group_by,
    clause_where,
    clause_where_combination,
    clause_order_by,
)

# processing data
from geoformat_lib.processing.data.field_statistics import field_statistics

# geoformat driver
from geoformat_lib.driver.geojson_driver import (
    geojson_to_geolayer,
    geolayer_to_geojson
)
from geoformat_lib.driver.postgresql_driver import geolayer_to_postgres

__version__ = 20210720


def version(version_value=__version__, verbose=True):
    version_dev = 'Alpha'
    if verbose:
        return '{version_dev} version {version_number}'.format(version_dev=version_dev, version_number=version_value)
    else:
        return version_value




import datetime

from tests.test_all import test_function

from geoformat_lib.conversion.metadata_conversion import (
    geometries_scan_to_geometries_metadata,
    fields_scan_to_fields_metadata,
    reorder_metadata_field_index_after_field_drop
)

geometries_scan_to_geometries_metadata_parameters = {
    0: {
        "geometry_scan": {'type': {'Polygon', 'MultiPolygon'}, "extent": None},
        "extent": False,
        "crs": None,
        "return_value": {'type': {'Polygon', 'MultiPolygon'}}
    },
    1: {
        "geometry_scan": {'type': {'Polygon', 'MultiPolygon'}, "extent": None},
        "extent": True,
        "crs": None,
        "return_value":  {'type': {'Polygon', 'MultiPolygon'}}
    },
    2: {
        "geometry_scan": {'type': None, "extent": None},
        "extent": False,
        "crs": None,
        "return_value": None
    },
    3: {
        "geometry_scan": {'type': None, "extent": None},
        "extent": True,
        "crs": None,
        "return_value": None
    },
    4: {
        "geometry_scan": {'type': {'Polygon'}, "extent": None},
        "crs": None,
        "extent": True,
        "return_value":  {'type': {'Polygon'}}
    },
    5: {
        "geometry_scan": {
            'type': {
                'Point',
                'LineString',
                'Polygon',
                'MultiPoint',
                'MultiLineString',
                'MultiPolygon',
                'GeometryCollection'
            },
            "extent": None
        },
        "crs": None,
        "extent": True,
        "return_value": {
            'type': {
                'Point',
                'LineString',
                'Polygon',
                'MultiPoint',
                'MultiLineString',
                'MultiPolygon',
                'GeometryCollection'
            }
         }
    },
    6: {
        "geometry_scan": {'type': {'Polygon', 'MultiPolygon'}, "extent": None},
        "crs": None,
        "extent": True,
        "return_value": {'type': {'Polygon', 'MultiPolygon'}}
    },
    7: {
        "geometry_scan": {'type': {'Polygon', 'MultiPolygon'}, "extent": (124277.0, 6050136.0, 1242213.0, 7110430.0)},
        "crs": None,
        "extent": True,
        "return_value": {'type': {'Polygon', 'MultiPolygon'}, "extent": (124277.0, 6050136.0, 1242213.0, 7110430.0)}
    },
    8: {
        "geometry_scan": {'type': None, "extent": None},
        "crs": None,
        "extent": True,
        "return_value": None
    },
    9: {
        "geometry_scan": {'type': {'Polygon'}, "extent": (382932.0, 6743442.0, 790134.0, 6997000.0)},
        "crs": None,
        "extent": True,
        "return_value":  {'type': {'Polygon'}, "extent": (382932.0, 6743442.0, 790134.0, 6997000.0)}
    },
    10: {
        "geometry_scan": {'type': {'Point', 'LineString', 'Polygon', 'MultiPoint', 'MultiLineString', 'MultiPolygon',
                                   'GeometryCollection'}, "extent": (-157.97, -34.29, 35.12, 77.95)},
        "crs": None,
        "extent": True,
        "return_value": {'type': {'Point', 'LineString', 'Polygon', 'MultiPoint', 'MultiLineString', 'MultiPolygon',
                                  'GeometryCollection'}, "extent": (-157.97, -34.29, 35.12, 77.95)}
    },
    11: {
        "geometry_scan": {'type': {'Polygon', 'MultiPolygon'}, "extent": (124277.0, 6050136.0, 1242213.0, 7110430.0)},
        "extent": True,
        "crs": 2154,
        "return_value": {'type': {'Polygon', 'MultiPolygon'},
                         "extent": (124277.0, 6050136.0, 1242213.0, 7110430.0),
                         "crs": 2154}
    },
    12: {
        "geometry_scan": {'type': {'Point', 'LineString', 'Polygon', 'MultiPoint', 'MultiLineString', 'MultiPolygon',
                                   'GeometryCollection'}, "extent": (-157.97, -34.29, 35.12, 77.95)},
        "extent": True,
        "crs": 4326,
        "return_value": {'type': {'Point', 'LineString', 'Polygon', 'MultiPoint', 'MultiLineString', 'MultiPolygon',
                                  'GeometryCollection'},
                         "extent": (-157.97, -34.29, 35.12, 77.95),
                         "crs": 4326}
    }
}

fields_scan_to_fields_metadata_parameters = {
    0: {
        "fields_scan": {
            "field_integer": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": float,
                "field_width_str": 6,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 0,
                "field_recast": False,
                "native_type": {float},
                "str_is_always_hexadecimal": None,
                "force_type": False,
                'field_width_float': 4,
                'width_before_comma': 4,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_integer_list": {
                "values_in_list": True,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 12,
                "field_width_list": 13,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 1,
                "field_recast": True,
                "native_type": {str, int, float},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 4,
                'width_before_comma': 4,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_real": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 7,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 2,
                "field_recast": True,
                "native_type": {str, int},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_real_list": {
                "values_in_list": True,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 19,
                "field_width_list": 6,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 3,
                "field_recast": True,
                "native_type": {str, int},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_string": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 26,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_index": 4,
                "field_recast": True,
                "native_type": {str, datetime.datetime},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_date": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.date,
                "field_width_str": 10,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_index": 5,
                "field_recast": False,
                "native_type": {datetime.date},
                "str_is_always_hexadecimal": None,
                "force_type": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_time": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 15,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 6,
                "field_recast": True,
                "native_type": {str, datetime.time},
                "str_is_always_hexadecimal": False,
                "force_type": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 201,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 7,
                "field_recast": True,
                "native_type": {str, bytes},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_boolean": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 5,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 8,
                "field_recast": True,
                "native_type": {bool, int},
                "str_is_always_hexadecimal": None,
                "force_type": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_none": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 4,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_index": 9,
                "field_recast": False,
                "native_type": {str},
                "str_is_always_hexadecimal": False,
                "force_type": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_string_list": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 16,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_index": 10,
                "field_recast": False,
                "native_type": {str},
                "str_is_always_hexadecimal": False,
                "force_type": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_datetime": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.datetime,
                "field_width_str": 26,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_index": 11,
                "field_recast": False,
                "native_type": {datetime.datetime},
                "str_is_always_hexadecimal": None,
                "force_type": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
        },
        "return_value": {
            "field_integer": {"type": "Real", "width": 4, "precision": 0, "index": 0},
            "field_integer_list": {"type": "String", "width": 13, "index": 1},
            "field_real": {"type": "String", "width": 7, "index": 2},
            "field_real_list": {"type": "String", "width": 19, "index": 3},
            "field_string": {"type": "String", "width": 26, "index": 4},
            "field_date": {"type": "Date", "index": 5},
            "field_time": {"type": "String", "width": 15, "index": 6},
            "field_binary": {"type": "String", "width": 201, "index": 7},
            "field_boolean": {"type": "String", "width": 5, "index": 8},
            "field_none": {"type": "String", "width": 4, "index": 9},
            "field_string_list": {"type": "String", "width": 16, "index": 10},
            "field_datetime": {"type": "DateTime", "index": 11},
        },
    },
    1: {
        "fields_scan": {
            "field_integer": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": float,
                "field_width_str": 6,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": False,
                "native_type": {float},
                "str_is_always_hexadecimal": None,
                "force_type": False,
                'field_width_float': 4,
                'width_before_comma': 4,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_integer_list": {
                "values_in_list": True,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 12,
                "field_width_list": 13,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, int, float},
                "str_is_always_hexadecimal": False,
                "force_type": False,
                'field_width_float': 4,
                'width_before_comma': 4,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_real": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 7,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, int},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_real_list": {
                "values_in_list": True,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 19,
                "field_width_list": 6,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, int},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_string": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 26,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, datetime.datetime},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_date": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.date,
                "field_width_str": 10,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_recast": False,
                "native_type": {datetime.date},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_time": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 15,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, datetime.time},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 201,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {str, bytes},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_boolean": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 5,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_recast": True,
                "native_type": {bool, int},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_none": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 4,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_recast": False,
                "native_type": {str},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_string_list": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 16,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_recast": False,
                "native_type": {str},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_datetime": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.datetime,
                "field_width_str": 26,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": True,
                "field_recast": False,
                "native_type": {datetime.datetime},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
        },
        "return_value": {
            "field_integer": {"type": "Real", "width": 4, "precision": 0},
            "field_integer_list": {"type": "String", "width": 13},
            "field_real": {"type": "String", "width": 7},
            "field_real_list": {"type": "String", "width": 19},
            "field_string": {"type": "String", "width": 26},
            "field_date": {"type": "Date"},
            "field_time": {"type": "String", "width": 15},
            "field_binary": {"type": "String", "width": 201},
            "field_boolean": {"type": "String", "width": 5},
            "field_none": {"type": "String", "width": 4},
            "field_string_list": {"type": "String", "width": 16},
            "field_datetime": {"type": "DateTime"},
        },
    },
    2: {
        "fields_scan": {
            "field_integer": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": int,
                "field_width_str": 3,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 0,
                "field_recast": False,
                "native_type": {int},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                'field_width_float': 3,
                'width_before_comma': 3,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_integer_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": int,
                "field_width_str": 4,
                "field_width_list": 12,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 1,
                "field_recast": False,
                "native_type": {int},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                'field_width_float': 4,
                'width_before_comma': 4,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_real": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": int,
                "field_width_str": 6,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 2,
                "field_recast": True,
                "native_type": {float},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                'field_width_float': 4,
                'width_before_comma': 4,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_real_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": float,
                "field_width_str": 10,
                "field_width_list": 24,
                "field_precision": 5,
                "none_value": False,
                "not_none_value": True,
                "field_index": 3,
                "field_recast": False,
                "native_type": {float},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                'field_width_float': 10,
                'width_before_comma': 5,
                'width_after_comma': 5,
                "field_delete": False
            },
            "field_string": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 5,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 4,
                "field_recast": False,
                "native_type": {str},
                "force_type": True,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_string_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": str,
                "field_width_str": 8,
                "field_width_list": 23,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 5,
                "field_recast": False,
                "native_type": {str},
                "force_type": True,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_none": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 6,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": True
            },
            "field_date": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.date,
                "field_width_str": 10,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 7,
                "field_recast": False,
                "native_type": {datetime.date},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_time": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.time,
                "field_width_str": 15,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 8,
                "field_recast": False,
                "native_type": {datetime.time},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_datetime": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.datetime,
                "field_width_str": 26,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 9,
                "field_recast": False,
                "native_type": {datetime.datetime},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bytes,
                "field_width_str": 201,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 10,
                "field_recast": False,
                "native_type": {bytes},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_boolean": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bool,
                "field_width_str": 4,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 11,
                "field_recast": False,
                "native_type": {bool},
                "force_type": True,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
        },
        "return_value": {
            "field_integer": {"type": "Integer", "index": 0},
            "field_integer_list": {"type": "IntegerList", "index": 1},
            "field_real": {"type": "Integer", "index": 2},
            "field_real_list": {"type": "RealList", "width": 10, "precision": 5, "index": 3},
            "field_string": {"type": "String", "width": 5, "index": 4},
            "field_string_list": {"type": "StringList", "width": 8, "index": 5},
            "field_date": {"type": "Date", "index": 6},
            "field_time": {"type": "Time", "index": 7},
            "field_datetime": {"type": "DateTime", "index": 8},
            "field_binary": {"type": "Binary", "index": 9},
            "field_boolean":  {"type": "Boolean", "index": 10}
        }
    },
    3: {
        "fields_scan": {
            "field_integer": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": int,
                "field_width_str": 3,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 0,
                "field_recast": False,
                "native_type": {int},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_integer_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": int,
                "field_width_str": 4,
                "field_width_list": 12,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 1,
                "field_recast": False,
                "native_type": {int},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_real": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": float,
                "field_width_str": 6,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 2,
                "field_recast": False,
                "native_type": {float},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 4,
                'width_before_comma': 4,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_real_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": float,
                "field_width_str": 10,
                "field_width_list": 24,
                "field_precision": 5,
                "none_value": False,
                "not_none_value": True,
                "field_index": 3,
                "field_recast": False,
                "native_type": {float},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 10,
                'width_before_comma': 5,
                'width_after_comma': 5,
                "field_delete": False
            },
            "field_string": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 5,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 4,
                "field_recast": False,
                "native_type": {str},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_string_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": str,
                "field_width_str": 8,
                "field_width_list": 23,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 5,
                "field_recast": False,
                "native_type": {str},
                "force_type": False,
                "str_is_always_hexadecimal": False,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_none": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 6,
                "field_recast": False,
                "native_type": set(),
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": True
            },
            "field_date": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.date,
                "field_width_str": 10,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 7,
                "field_recast": False,
                "native_type": {datetime.date},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_time": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.time,
                "field_width_str": 15,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 8,
                "field_recast": False,
                "native_type": {datetime.time},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_datetime": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.datetime,
                "field_width_str": 26,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 9,
                "field_recast": False,
                "native_type": {datetime.datetime},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bytes,
                "field_width_str": 201,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 10,
                "field_recast": False,
                "native_type": {bytes},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_boolean": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bool,
                "field_width_str": 4,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 11,
                "field_recast": False,
                "native_type": {bool},
                "force_type": False,
                "str_is_always_hexadecimal": None,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
        },
        "return_value": {
            "field_integer": {"type": "Integer", "index": 0},
            "field_integer_list": {"type": "IntegerList", "index": 1},
            "field_real": {"type": "Real", "width": 4, "precision": 0, "index": 2},
            "field_real_list": {"type": "RealList", "width": 10, "precision": 5, "index": 3},
            "field_string": {"type": "String", "width": 5, "index": 4},
            "field_string_list": {"type": "StringList", "width": 8, "index": 5},
            "field_date": {"type": "Date", "index": 6},
            "field_time": {"type": "Time", "index": 7},
            "field_datetime": {"type": "DateTime", "index": 8},
            "field_binary": {"type": "Binary", "index": 9},
            "field_boolean":  {"type": "Boolean", "index": 10}
        }
    },
    4: {
        "fields_scan": {
            "field_none": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 0,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": True
            },
            "field_integer": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": int,
                "field_width_str": 3,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 1,
                "field_recast": False,
                "native_type": {int},
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_integer_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": int,
                "field_width_str": 4,
                "field_width_list": 12,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 2,
                "field_recast": False,
                "native_type": {int},
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_real": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": int,
                "field_width_str": 6,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 3,
                "field_recast": True,
                "native_type": {float},
                "force_type": True,
                'field_width_float': 4,
                'width_before_comma': 4,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_real_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": float,
                "field_width_str": 10,
                "field_width_list": 24,
                "field_precision": 5,
                "none_value": False,
                "not_none_value": True,
                "field_index": 4,
                "field_recast": False,
                "native_type": {float},
                "force_type": True,
                'field_width_float': 10,
                'width_before_comma': 5,
                'width_after_comma': 5,
                "field_delete": False
            },
            "field_string": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": str,
                "field_width_str": 5,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 5,
                "field_recast": False,
                "native_type": {str},
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_string_list": {
                "values_in_list": True,
                "values_out_list": False,
                "field_list": True,
                "field_type": str,
                "field_width_str": 8,
                "field_width_list": 23,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 6,
                "field_recast": False,
                "native_type": {str},
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_none_1": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 7,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": True
            },
            "field_date": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.date,
                "field_width_str": 10,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 8,
                "field_recast": False,
                "native_type": {datetime.date},
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_time": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.time,
                "field_width_str": 15,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 9,
                "field_recast": False,
                "native_type": {datetime.time},
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_none_2": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 10,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": True
            },
            "field_none_3": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 11,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": True
            },
            "field_datetime": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": datetime.datetime,
                "field_width_str": 26,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 12,
                "field_recast": False,
                "native_type": {datetime.datetime},
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bytes,
                "field_width_str": 201,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 13,
                "field_recast": False,
                "native_type": {bytes},
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_boolean": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bool,
                "field_width_str": 4,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 14,
                "field_recast": False,
                "native_type": {bool},
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            },
            "field_none_4": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": None,
                "field_width_str": 0,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": True,
                "not_none_value": False,
                "field_index": 15,
                "field_recast": False,
                "native_type": set(),
                "force_type": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": True
            },
        },
        "return_value": {
            "field_integer": {"type": "Integer", "index": 0},
            "field_integer_list": {"type": "IntegerList", "index": 1},
            "field_real": {"type": "Integer", "index": 2},
            "field_real_list": {"type": "RealList", "width": 10, "precision": 5, "index": 3},
            "field_string": {"type": "String", "width": 5, "index": 4},
            "field_string_list": {"type": "StringList", "width": 8, "index": 5},
            "field_date": {"type": "Date", "index": 6},
            "field_time": {"type": "Time", "index": 7},
            "field_datetime": {"type": "DateTime", "index": 8},
            "field_binary": {"type": "Binary", "index": 9},
            "field_boolean":  {"type": "Boolean", "index": 10}
        }
    },
    5: {
        "fields_scan": {
            "field_binary": {
                "values_in_list": False,
                "values_out_list": True,
                "field_list": False,
                "field_type": bytes,
                "field_width_str": 201,
                "field_width_list": None,
                "field_precision": 0,
                "none_value": False,
                "not_none_value": True,
                "field_index": 0,
                "field_recast": True,
                "native_type": {str, bytes},
                "force_type": True,
                "str_is_always_hexadecimal": True,
                'field_width_float': 0,
                'width_before_comma': 0,
                'width_after_comma': 0,
                "field_delete": False
            }
        },
        "return_value": {
            "field_binary": {"type": "Binary", 'index': 0}
        }
    }
}

reorder_metadata_field_index_after_field_drop_parameters = {
    0: {
        "fields_metadata": {
            'NOM_DEPT': {'type': 'String', 'width': 10, 'index': 1}
        },
        "reorder_field_index": 0,
        "return_value": {'NOM_DEPT': {'type': 'String', 'width': 10, 'index': 0}}
    },
    1: {
        "fields_metadata": {'CODE_DEPT': {'type': 'String', 'width': 2, 'index': 0}},
        "reorder_field_index": 1,
        "return_value": {'CODE_DEPT': {'type': 'String', 'width': 2, 'index': 0}}
    },
    2: {
        "fields_metadata": {
            'field_integer': {'type': 'Integer', 'index': 0},
            'field_integer_list': {'type': 'IntegerList', 'index': 1},
            'field_real': {'type': 'Real', 'width': 6, 'precision': 0, 'index': 2},
            'field_real_list': {'type': 'RealList', 'width': 10, 'precision': 5, 'index': 3},
            'field_date': {'type': 'Date', 'index': 6},
            'field_time': {'type': 'Time', 'index': 7},
            'field_datetime': {'type': 'DateTime', 'index': 8},
            'field_binary': {'type': 'Binary', 'index': 9},
            'field_boolean': {'type': 'Boolean', 'index': 10}
        },
        "reorder_field_index": [4, 5],
        "return_value": {
            'field_integer': {'type': 'Integer', 'index': 0},
            'field_integer_list': {'type': 'IntegerList', 'index': 1},
            'field_real': {'type': 'Real', 'width': 6, 'precision': 0, 'index': 2},
            'field_real_list': {'type': 'RealList', 'width': 10, 'precision': 5, 'index': 3},
            'field_date': {'type': 'Date', 'index': 4},
            'field_time': {'type': 'Time', 'index': 5},
            'field_datetime': {'type': 'DateTime', 'index': 6},
            'field_binary': {'type': 'Binary', 'index': 7},
            'field_boolean': {'type': 'Boolean', 'index': 8}
        }
    },
    3: {
        "fields_metadata": {
            'field_real': {'type': 'Real', 'width': 6, 'precision': 0, 'index': 2},
            'field_real_list': {'type': 'RealList', 'width': 10, 'precision': 5, 'index': 3},
            'field_string': {'type': 'String', 'width': 5, 'index': 4},
            'field_string_list': {'type': 'StringList', 'width': 8, 'index': 5},
            'field_date': {'type': 'Date', 'index': 6},
            'field_time': {'type': 'Time', 'index': 7},
            'field_datetime': {'type': 'DateTime', 'index': 8},
            'field_binary': {'type': 'Binary', 'index': 9},
            'field_boolean': {'type': 'Boolean', 'index': 10}
        },
        "reorder_field_index": [0, 1],
        "return_value": {
            'field_real': {'type': 'Real', 'width': 6, 'precision': 0, 'index': 0},
            'field_real_list': {'type': 'RealList', 'width': 10, 'precision': 5, 'index': 1},
            'field_string': {'type': 'String', 'width': 5, 'index': 2},
            'field_string_list': {'type': 'StringList', 'width': 8, 'index': 3},
            'field_date': {'type': 'Date', 'index': 4},
            'field_time': {'type': 'Time', 'index': 5},
            'field_datetime': {'type': 'DateTime', 'index': 6},
            'field_binary': {'type': 'Binary', 'index': 7},
            'field_boolean': {'type': 'Boolean', 'index': 8}
        }
    },
    4: {
        "fields_metadata": {
            'field_integer': {'type': 'Integer', 'index': 0},
            'field_integer_list': {'type': 'IntegerList', 'index': 1},
            'field_real': {'type': 'Real', 'width': 6, 'precision': 0, 'index': 2},
            'field_real_list': {'type': 'RealList', 'width': 10, 'precision': 5, 'index': 3},
            'field_string': {'type': 'String', 'width': 5, 'index': 4},
            'field_string_list': {'type': 'StringList', 'width': 8, 'index': 5},
            'field_date': {'type': 'Date', 'index': 6},
            'field_time': {'type': 'Time', 'index': 7}
        },
        "reorder_field_index": [8, 9, 10],
        "return_value":  {
            'field_integer': {'type': 'Integer', 'index': 0},
            'field_integer_list': {'type': 'IntegerList', 'index': 1},
            'field_real': {'type': 'Real', 'width': 6, 'precision': 0, 'index': 2},
            'field_real_list': {'type': 'RealList', 'width': 10, 'precision': 5, 'index': 3},
            'field_string': {'type': 'String', 'width': 5, 'index': 4},
            'field_string_list': {'type': 'StringList', 'width': 8, 'index': 5},
            'field_date': {'type': 'Date', 'index': 6},
            'field_time': {'type': 'Time', 'index': 7},
        }
    },
}


def test_all():
    # geometries_scan_to_metadata
    print(test_function(geometries_scan_to_geometries_metadata, geometries_scan_to_geometries_metadata_parameters))

    # fields_scan_to_fields_metadata
    print(test_function(fields_scan_to_fields_metadata, fields_scan_to_fields_metadata_parameters))

    # reorder_metadata_field_index_after_field_drop
    print(test_function(
        reorder_metadata_field_index_after_field_drop,
        reorder_metadata_field_index_after_field_drop_parameters
    )
    )


if __name__ == '__main__':
    test_all()
